//	TrayIcon.cpp - Where the app does all its work.
//	----------------------------------------------------------------------------
//	This file is part of OSC2VST, a simple OSC-aware VSTi host.
//	Copyright (c) 2009 Niall Moody.
//
//	This program is free software: you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation, either version 3 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program.  If not, see <http://www.gnu.org/licenses/>.
//	----------------------------------------------------------------------------

#include "TrayIcon.h"
#include "Images.h"
#include "PluginWindow.h"
#include "NativeDialog.h"
#include "ParameterList.h"
#include "AudioSettings.h"
#include "NiallsOSCLib/OSCBundle.h"
#include "NiallsOSCLib/OSCMessage.h"

using namespace std;

//------------------------------------------------------------------------------
TrayIcon::TrayIcon(const String& commandLine):
Thread(T("OSC2VST Listening Thread")),
icon(0),
plugin(0),
pluginWindow(0),
newPort(5678),
laf(0)
{

	icon = ImageCache::getFromMemory(Images::icon_png, Images::icon_pngSize);

	if(icon)
		setIconImage(*icon);

	//Setup alternative LookAndFeel.
	LookAndFeel::setDefaultLookAndFeel(laf = new BranchesLAF());

	//Set port to listen on.
	socket.setPort(5678);

	//Make sure we'll recognise VST plugins.
	AudioPluginFormatManager::getInstance()->addDefaultFormats();

	//Test against odd strings being passed in on the command line???
	if(commandLine.isNotEmpty())
		loadPlugin(commandLine);
	else
	{
		setIconTooltip(T("OSC2VST: <no plugin loaded>"));

		/*AlertWindow::showMessageBox(AlertWindow::WarningIcon,
                                    T("File error"),
                                    T("No plugin path selected."));*/
	}

	//Start the audio.
	deviceManager.refreshDeviceList();
	deviceManager.setAudioCallback(&pluginPlayer);
	if(plugin)
		initialiseAudio();

	//Start the OSC thread.
	startThread();
	//Start the timer to watch for new plugins being loaded.
	startTimer(30);
}

//------------------------------------------------------------------------------
TrayIcon::~TrayIcon()
{
	stopThread(250);
	deviceManager.closeAudioDevice();

	//Save audio settings.
	saveAudioSettings();
	if(0)
	{
		XmlElement *settingsFile;
		File audioSettings = File::getSpecialLocation(File::userApplicationDataDirectory).getChildFile(T("OSC2VST"));

		if(!audioSettings.exists())
			audioSettings.createDirectory();
		audioSettings = audioSettings.getChildFile(T("AudioSettings.xml"));

		settingsFile = deviceManager.createStateXml();
		settingsFile->writeToFile(audioSettings, T(""));

		delete settingsFile;
	}

	if(pluginWindow)
		delete pluginWindow;
	if(plugin)
	{
		plugin->releaseResources();
		delete plugin;
	}

	ImageCache::release(icon);

	LookAndFeel::setDefaultLookAndFeel(0);
	delete laf;
}

//------------------------------------------------------------------------------
void TrayIcon::mouseUp(const MouseEvent& e)
{
	if(e.mods.isRightButtonDown())
	{
		PopupMenu m;
		if(plugin)
		{
			m.addItem(1, "Audio Settings");
			m.addSeparator();
			m.addItem(2, "Display Plugin Editor");
			m.addItem(3, "List of Plugin Parameters");
		}
		else //Don't let the user break things if there's no plugin to generate
		{	 //any audio.
			m.addItem(1, "Audio Settings", false);
			m.addSeparator();
			m.addItem(2, "Display Plugin Editor", false);
			m.addItem(3, "List of Plugin Parameters", false);
		}
		m.addSeparator();
		m.addItem(4, "Exit");

		const int result = m.show();

		switch(result)
		{
			case 1:
				{
					/*AudioDeviceSelectorComponent *win = new AudioDeviceSelectorComponent(deviceManager,
																						 0,
																						 0,
																						 2,
																						 2,
																						 false,
																						 false);*/
					AudioSettingsComponent *win = new AudioSettingsComponent(&deviceManager, this);
					win->setSize(350, 174);

					NativeDialog::showNonModalDialog(T("Audio Settings"),
												  win,
												  0,
												  Colour(0xFFEEECE1));
				}
				break;
			case 2:
				if(!pluginWindow)
					pluginWindow = new PluginWindow(plugin, this);
				break;
			case 3:
				{
					Viewport *viewport = new Viewport();
					ParameterList *win = new ParameterList(plugin);

					viewport->setSize(400, 400);
					viewport->setViewedComponent(win);

					NativeDialog::showNonModalDialog(T("Available OSC Addresses"),
												     viewport,
												     0,
												     Colour(0xFFEEECE1));
				}
				break;
			case 4:
				JUCEApplication::quit();
				break;
		}
	}
}

//------------------------------------------------------------------------------
void TrayIcon::run()
{
	char *buffer;
	int32_t size;
	String address;

	socket.bindSocket(true);

	while(!threadShouldExit())
	{
		//Check if the port's been changed.
		if(newPort != socket.getPort())
		{
			socket.setPort(newPort);
			socket.bindSocket(true);
		}

		buffer = socket.getData(size);

		if(size > 0)
		{
			if(OSC::Bundle::isBundle(buffer, size))
			{
				OSC::Bundle bundle(buffer, size);

				handleBundle(bundle);
			}
			else if(OSC::Message::isMessage(buffer, size))
			{
				OSC::Message mess(buffer, size);

				handleMessage(mess);
			}
		}
	}
}

//------------------------------------------------------------------------------
void TrayIcon::timerCallback()
{
	if(newPlugin.isNotEmpty() && (newPlugin != T("****UNLOAD****")))
	{
		deviceManager.closeAudioDevice();
		loadPlugin(newPlugin);
		if(plugin)
			initialiseAudio();

		newPlugin = String::empty;
	}
	else if(newPlugin == T("****UNLOAD****"))
	{
		if(plugin)
		{
			if(pluginWindow)
			{
				delete pluginWindow;
				pluginWindow = 0;
			}
			parameters.clear();
			pluginPlayer.setProcessor(0);
			plugin->releaseResources();
			delete plugin;
			plugin = 0;

			setIconTooltip(T("OSC2VST: <no plugin loaded>"));
		}
	}
}

//------------------------------------------------------------------------------
void TrayIcon::setPort(int16_t p)
{
	newPort = p;
}

//------------------------------------------------------------------------------
void TrayIcon::editorClosed()
{
	pluginWindow = 0;
}

//------------------------------------------------------------------------------
void TrayIcon::loadPlugin(const String& path)
{
	int i;
	KnownPluginList tempList;
	OwnedArray<PluginDescription> tempArr;

	tempList.scanAndAddFile(File(path), false, tempArr);
	if(tempArr.size() > 0)
	{
		String err;

		//Clear previous settings if this isn't the first time we've loaded a
		//plugin.
		if(plugin)
		{
			if(pluginWindow)
			{
				delete pluginWindow;
				pluginWindow = 0;
			}
			parameters.clear();
			pluginPlayer.setProcessor(0);
			plugin->releaseResources();
			delete plugin;
			plugin = 0;
		}

		plugin = AudioPluginFormatManager::getInstance()->createPluginInstance(*(tempArr[0]), err);

		if(plugin)
		{
			String tempstr;

			pluginPlayer.setProcessor(plugin);
			tempstr << T("OSC2VST: ") << tempArr[0]->name;
			setIconTooltip(tempstr);

			//Fill out the parameters map.
			for(i=0;i<plugin->getNumParameters();++i)
			{
				String tempParam;
				String tempstr;

				tempstr = plugin->getParameterName(i);
				tempstr = tempstr.removeCharacters(T(" #*,/?[]{}"));
				tempParam << T("/") << tempstr;

				//Watch for plugins which give multiple parameters the same
				//name.  Should figure out a better solution than just ignoring
				//them...
				if(parameters.find((const char *)tempParam) == parameters.end())
				{
					parameters.insert(parameters.end(),
									  make_pair(string((const char *)tempParam), i));
				}
			}

			plugin->prepareToPlay(44100.0, 4028);
		}
		else
		{
			AlertWindow::showMessageBox(AlertWindow::WarningIcon,
                                        T("Could not load plugin"),
                                        err);
		}
	}
	else
	{
		AlertWindow::showMessageBox(AlertWindow::WarningIcon,
                                    T("File error"),
                                    T("Could not determine a valid plugin from the file path passed in."));
	}
}

//------------------------------------------------------------------------------
void TrayIcon::handleMessage(OSC::Message& message)
{
	string address;

	address = message.getAddress();

	if(address == "/quit")
		JUCEApplication::quit();
	else if(address == "/loadPlugin")
	{
		if(message.getNumStrings() > 0)
		{
			if(message.getString(0) != "")
				newPlugin = message.getString(0).c_str();
			else
				newPlugin = T("****UNLOAD****");
		}
	}
	else if(address == "/unloadPlugin")
		newPlugin = T("****UNLOAD****");
	//Just make sure plugin exists before we try and set any parameters.
	else if(plugin)
	{
		if(address == "/noteOn")
		{
			MidiMessageCollector& collector = pluginPlayer.getMidiMessageCollector();

			if((message.getNumInts() > 0) && (message.getNumFloats() > 0))
			{
				MidiMessage midi = MidiMessage::noteOn(1, message.getInt(0), message.getFloat(0));
				midi.setTimeStamp(Time::getMillisecondCounter()/1000.0);

				collector.addMessageToQueue(midi);
			}
		}
		else if(address == "/noteOff")
		{
			MidiMessageCollector& collector = pluginPlayer.getMidiMessageCollector();

			if(message.getNumInts() > 0)
			{
				MidiMessage midi = MidiMessage::noteOff(1, message.getInt(0));
				midi.setTimeStamp(Time::getMillisecondCounter()/1000.0);

				collector.addMessageToQueue(midi);
			}
		}
		else if(address == "/MIDI")
		{
			MidiMessageCollector& collector = pluginPlayer.getMidiMessageCollector();

			if(message.getNumInts() > 2)
			{
				MidiMessage midi(message.getInt(0),
								 message.getInt(1),
								 message.getInt(2),
								 Time::getMillisecondCounter()/1000.0);

				collector.addMessageToQueue(midi);
			}
		}
		else
		{
			map<string, int>::iterator it = parameters.find(address);

			if((it != parameters.end()) && (message.getNumFloats() > 0))
				plugin->setParameter(it->second, message.getFloat(0));
		}
	}
}

//------------------------------------------------------------------------------
void TrayIcon::handleBundle(OSC::Bundle& bundle)
{
	long i;

	for(i=0;i<bundle.getNumMessages();++i)
		handleMessage(*(bundle.getMessage(i)));

	for(i=0;i<bundle.getNumBundles();++i)
		handleBundle(*(bundle.getBundle(i)));
}

//------------------------------------------------------------------------------
File TrayIcon::getAudioSettingsFile() const
{
	File retval = File::getSpecialLocation(File::currentExecutableFile).getParentDirectory();

	//First check if the file exists in the application's folder.
	retval = retval.getChildFile(T("AudioSettings.xml"));
	if(!retval.existsAsFile())
	{
		//If it doesn't, it must be in the user's AppData folder.
		retval = File::getSpecialLocation(File::userApplicationDataDirectory).getChildFile(T("OSC2VST"));

		if(!retval.exists())
			retval.createDirectory();
		retval = retval.getChildFile(T("AudioSettings.xml"));
	}

	return retval;
}

//------------------------------------------------------------------------------
void TrayIcon::initialiseAudio()
{
	File settingsFile = getAudioSettingsFile();

	if(settingsFile.existsAsFile())
	{
		XmlDocument settingsDoc(settingsFile);
		XmlElement *rootEl = settingsDoc.getDocumentElement();
		XmlElement *settingsEl;

		if(rootEl)
		{
			if(rootEl->hasTagName(T("OSC2VST")))
			{
				//Get audio settings.
				settingsEl = rootEl->getChildByName(T("DEVICESETUP"));
				if(settingsEl)
					deviceManager.initialise(0, 2, settingsEl, true);
				else
					deviceManager.initialise(0, 2, 0, true);

				//Get OSC port.
				settingsEl = rootEl->getChildByName(T("OSCPort"));
				if(settingsEl)
					newPort = settingsEl->getIntAttribute(T("value"));
			}
			else
				deviceManager.initialise(0, 2, 0, true);

			delete rootEl;
		}
		else
			deviceManager.initialise(0, 2, 0, true);
	}
	else
		deviceManager.initialise(0, 2, 0, true);
}

//------------------------------------------------------------------------------
void TrayIcon::saveAudioSettings()
{
	File settingsFile = getAudioSettingsFile();
	XmlElement *rootEl = new XmlElement(T("OSC2VST"));
	XmlElement *deviceSettings;
	XmlElement *oscPort = new XmlElement(T("OSCPort"));

	deviceSettings = deviceManager.createStateXml();
	rootEl->addChildElement(deviceSettings);
	oscPort->setAttribute(T("value"), socket.getPort());
	rootEl->addChildElement(oscPort);
	rootEl->writeToFile(settingsFile, T(""));

	delete rootEl;
}
